"""
Обработка QuerySet.
"""
import sqlparse
from django.shortcuts import _get_queryset


def quick_pagination(queryset, page, limit, max_limit=1000):
    """
    Функция быстрой паджинации для больших таблиц, не вызывающая count(*).
    """
    try:
        page = int(page)
    except ValueError:
        page = 1
    try:
        limit = int(limit)
        assert limit <= max_limit
    except ValueError:
        limit = 10
    except AssertionError:
        limit = max_limit
    offset = (page - 1) * limit
    queryset = queryset[offset:(offset + limit + 1)]
    result = queryset
    if len(result) > limit:
        has_next = True
        result = result[:limit]
    else:
        has_next = False
    return result, page, limit, has_next


def dictfetchall(cursor, aggregate=()):
    """
    Return all rows from a cursor as a dict.
    """
    columns = [col[0] for col in cursor.description]

    total = {col: 0 for col in aggregate}
    use_total = bool(total)
    total_keys = total.keys()

    def to_dict(row):
        d = dict(zip(columns, row))
        if use_total:
            for col in total_keys:
                total[col] += d[col]
        return d

    objects = [to_dict(row) for row in cursor.fetchall()]
    if total:
        return objects, total
    return objects


def get_object_or_none(klass, *args, **kwargs):
    """
    Возвращает объект или None, если объект не существует.

    klass может быть Model, Manager, или объектом QuerySet. Все остальные
    переданные параметры используются для запроса get().

    Замечание: Возвращает None, если найдено более одного объекта.
    """
    queryset = _get_queryset(klass)
    try:
        return queryset.get(*args, **kwargs)
    except queryset.model.DoesNotExist:
        return None
    except queryset.model.MultipleObjectsReturned:
        return None


def pretty_sql(queryset):
    """Форматирует QuerySet в человекочитаемый SQL."""
    query = queryset.query
    compiler = query.get_compiler(queryset.db)
    with compiler.connection.schema_editor() as editor:
        sql, params = compiler.as_sql()
        value = sql % tuple(editor.quote_value(p) for p in params)
    return sqlparse.format(value, reindent=True, keyword_case='upper')



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
"""
Вспомогательный функционал для обработки запросов.
"""
from django.contrib.auth import get_user_model


def get_ip_address(request, skip=None):
    meta = request.META
    ip = meta.get(
        'HTTP_X_REAL_IP', meta.get(
            'HTTP_X_FORWARDED_FOR', meta.get(
                'REMOTE_ADDR', ''
            )
        )
    )
    if ip and skip:
        if skip == 'local':
            skip = ('127.', '10.', '192.')
        for prefix in tuple(skip):
            if ip.startswith(prefix):
                return ''
    return ip


def get_cookies_domain2(request):
    """Возвращает домен второго уровня для кук."""
    host = request.get_host()
    domain = '.'.join(host.split('.')[-2:])
    if ':' in domain:
        domain = domain.split(':')[0]
    return domain


def get_user(request, user_id):
    user = request.user
    if user.id != user_id:
        user = get_user_model().objects.get(id=user_id)
    return user



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from django.core import validators
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _


@deconstructible
class PhoneValidator(validators.RegexValidator):
    regex = r'^\+(9[976]\d|8[987530]\d|6[987]\d|5[90]\d|42\d|3[875]\d|2[98654321]\d|9[8543210]|8[6421]|6[6543210]|5[87654321]|4[987654310]|3[9643210]|2[70]|7|1)\d{1,14}$'
    message = _('Введите правильный телефонный номер.')


@deconstructible
class WebmoneyIDValidator(validators.RegexValidator):
    regex = r'^\d{12}$'
    message = _('Введите правильный идентификтор Webmoney.')


@deconstructible
class WebmoneyWalletValidator(validators.RegexValidator):
    regex = r'^[R,Z]\d{12}$'
    message = _('Укажите правильный Webmoney-кошелёк.')


@deconstructible
class YandexMoneyWalletValidator(validators.RegexValidator):
    regex = r'^\d{11,16}$'
    message = _('Укажите правильный кошелёк от Яндекс-Денег.')


@deconstructible
class SteamIDValidator(validators.RegexValidator):
    regex = r'^\d{17}$'
    message = _('Введите правильный идентификтор Steam.')


@deconstructible
class SteamTradeLinkValidator(validators.RegexValidator):
    regex = r'^https://steamcommunity.com/tradeoffer/new/\?partner=[0-9]+&token=[a-zA-Z0-9_-]+$'
    message = _('Введите правильную ссылку на обмен в Steam.')


@deconstructible
class UUIDValidator(validators.RegexValidator):
    regex = r'^[a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}-?[89ab][a-f0-9]{3}-?[a-f0-9]{12}'
    message = _('Введите правильный UUID.')


@deconstructible
class BitcoinAddressValidator(validators.RegexValidator):
    regex = r'^[13][a-km-zA-HJ-NP-Z1-9]{25,34}$'
    message = _('Введите правильный платёжный Bitcoin-адрес.')



#
# Copyright (c) 2012, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from datetime import datetime
from django.utils.timezone import (
    now, utc, is_naive, get_current_timezone, get_default_timezone,
    activate, localtime,
)
from django.utils.dateparse import parse_datetime


def from_unixtime(timestamp=0, tzinfo=utc):
    """
    Преобразование POSIX времени в объект `datetime` с конечной временной зоной
    `tzinfo`. Заметьте, что POSIX время по стандарту всегда в UTC.
    """
    if not isinstance(timestamp, float):
        timestamp = float(timestamp)
    dt = datetime.fromtimestamp(timestamp, utc)
    if tzinfo == utc:
        return dt
    return dt.replace(tzinfo=tzinfo)


def from_javascript(timestamp=0, tzinfo=utc):
    """
    Получение времени из стандартного времени JavaScript (с милисекундами),
    например, такого: `var ms = new Date().getTime()`.
    Заметьте, что исходное время JavaScript по стандарту всегда в UTC, а
    параметр `tzinfo` служит для получения времени в конечном виде с
    необходимой временной зоной.
    """
    if not isinstance(timestamp, int):
        timestamp = int(timestamp)
    return from_unixtime(timestamp / 1000.0, tzinfo=tzinfo)


def datetime_round(dt=None, minute=True, hour=False, day=False):
    """
    Округление времени до секунды, минуты, часа или дня.
    По умолчанию до минуты.
        до дня:     rounded(day=True)
        до часа:    rounded(hour=True)
        до минуты:  rounded()
        до секунды: rounded(minute=False)

    Если нужно округлить заданное время,то оно передаётся в
    параметре `dt`.

    На вход можно подавать как naive, так и aware, результатом будет
    аналогичное.
    Если генерируется новое время, то результатом будет naive при
    settings.USE_TZ = False, иначе - aware.
    """
    if not isinstance(dt, datetime):
        dt = datetime_local()

    dt = dt.replace(microsecond=0)

    if day:
        dt = dt.replace(hour=0, minute=0, second=0)
    elif hour:
        dt = dt.replace(minute=0, second=0)
    elif minute:
        dt = dt.replace(second=0)

    return dt


def datetime_local(value=None):
    """
    Переводит значение в текущую зону локального времени.
    На вход можно подавать как naive, так и aware, результатом будет
    аналогичное.
    Если генерируется новое время, то результатом будет naive при
    settings.USE_TZ = False, иначе - aware.
    """
    if not value:
        value = now()
    if is_naive(value):
        return value
    return localtime(value)


def datetime_server(value=None):
    """
    Переводит значение в зону серверного времени.
    На вход можно подавать как naive, так и aware, результатом будет
    аналогичное.
    Если генерируется новое время, то результатом будет naive при
    settings.USE_TZ = False, иначе - aware.
    """
    if not value:
        value = datetime_local()

    if is_naive(value):
        return value

    tz_curr = get_current_timezone()
    tz_serv = get_default_timezone()

    if tz_curr != tz_serv:
        activate(tz_serv)
        value = localtime(value, tz_serv)
        activate(tz_curr)
    else:
        value = localtime(value, tz_curr)

    return value


def datetime_naive(value=None):
    """
    Переводит значение в текущую зону локального времени.
    Возвращает простое время.
    """
    return datetime_local(value).replace(tzinfo=None)


def datetime_aware(value=None):
    """
    Переводит значение (даже простое) в текущую зону локального времени.
    Возвращает aware c текущей временной зоной.
    """
    tzinfo = get_current_timezone()
    if value and is_naive(value):
        if hasattr(tzinfo, 'localize'):
            # This method is available for pytz time zones.
            return tzinfo.localize(value)
        return value.replace(tzinfo=tzinfo)
    return localtime(value, tzinfo)


def datetime_server_naive(value=None):
    """
    Переводит значение в зону серверного времени settings.TIME_ZONE.
    Возвращает простое время.
    """
    return datetime_server(value).replace(tzinfo=None)


def datetime_server_aware(value=None):
    """
    Переводит значение (даже простое) в зону серверного времени.
    Возвращает aware c временной зоной сервера.
    """
    tzinfo = get_default_timezone()
    if value and is_naive(value):
        if hasattr(tzinfo, 'localize'):
            # This method is available for pytz time zones.
            return tzinfo.localize(value)
        return value.replace(tzinfo=tzinfo)
    return localtime(value, tzinfo)


def server_date():
    """
    Возвращает текущую дату сервера по settings.TIME_ZONE.
    """
    return datetime_server_naive().date()


def server_time():
    """
    Возвращает текущее время сервера по settings.TIME_ZONE.
    """
    return datetime_server_naive().time()


def parse_datetime_to_naive(value):
    """
    Парсер строкового значения (naive или aware) в локальное naive.
    """
    value = parse_datetime(value)
    if value:
        return datetime_naive(value)
    return None


def parse_datetime_to_aware(value):
    """
    Парсер строкового значения (naive или aware) в локальное aware.
    """
    value = parse_datetime(value)
    if value:
        return datetime_aware(value)
    return None


def parse_datetime_to_server_naive(value):
    """
    Парсер строкового значения (naive или aware) в серверное naive.
    """
    value = parse_datetime(value)
    if value:
        return datetime_server_naive(value)
    return None


def parse_datetime_to_server_aware(value):
    """
    Парсер строкового значения (naive или aware) в серверное aware.
    """
    value = parse_datetime(value)
    if value:
        return datetime_server_aware(value)
    return None



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from datetime import datetime
from django import forms
from django.forms.utils import from_current_timezone
from django.utils.dateparse import parse_datetime


class IsoDateTimeField(forms.DateTimeField):
    """DateTimeField with support ISO format."""

    def prepare_value(self, value):
        if isinstance(value, datetime):
            value = value.isoformat()
        return value

    def to_python(self, value):
        # First parse from ISO format.
        if value and isinstance(value, str):
            result = parse_datetime(value)
            if result:
                return from_current_timezone(result)
        # If not ISO format then standard behavior.
        return super().to_python(value)



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
try:
    from django.db.models import JSONField as BaseJSONField
except ImportError:
    from django.contrib.postgres.fields.jsonb import JSONField as BaseJSONField
from djangokit.forms.postgres import JSONField as FormJSONField
from djangokit.utils.encoders import JSONEncoder


class JSONField(BaseJSONField):
    """Django PostgreSQL JSONField with custom encoder."""

    def __init__(self, *args, **kwargs):
        super(JSONField, self).__init__(*args, **kwargs)
        if not self.encoder:
            self.encoder = JSONEncoder

    def formfield(self, **kwargs):
        return super().formfield(**{
            'form_class': FormJSONField,
            **kwargs,
        })



#
# Copyright (c) 2021, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
import logging
from sqlparse import format

from django.conf import settings
from django.db import connection, reset_queries


def logging_queries(get_response):
    """
    Выводит в лог SQL-запросы на уровне DEBUG.
    """

    logger = logging.getLogger(__name__)
    REINDENT = getattr(settings, 'PRETTY_SQL_REINDENT', False)

    def pretty_sql(sql):
        return format(sql, reindent=REINDENT, keyword_case='upper')

    def middleware(request):
        if settings.DEBUG:
            reset_queries()
            start_queries = len(connection.queries)

            response = get_response(request)

            end_queries = len(connection.queries)

            queries = []
            total_time = 0
            for i, query in enumerate(connection.queries):
                time = query['time']
                total_time += float(time)
                queries.append(
                    f"QUERY #{i + 1}. {time} seconds:\n"
                    f"{pretty_sql(query['sql'])}"
                )

            logger.debug(
                '%s %s -- %d database queries at %f seconds.\n\n%s\n',
                request.method,
                request.path,
                end_queries - start_queries,
                total_time,
                '\n\n'.join(queries)
            )
        else:
            response = get_response(request)

        return response

    return middleware



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
import re
from django import template
from django.template.base import Node
from django.utils.functional import keep_lazy_text
from djangokit.utils.markdown import markdown_to_html

register = template.Library()


@keep_lazy_text
def strip_blanklines(value):
    return re.sub(r'\n\s*\n', '\n', str(value))


class BlanklineLessNode(Node):
    def __init__(self, nodelist):
        self.nodelist = nodelist

    def render(self, context):
        func = strip_blanklines
        return func(self.nodelist.render(context).strip())


@register.tag
def blanklineless(parser, token):
    """
    Remove blank lines.

    Example usage::

        {% blanklineless %}
        First Line

        Second line
        {% endblanklineless %}

    This example returns this text::

        First Line
        Second line

    """
    nodelist = parser.parse(('endblanklineless',))
    parser.delete_first_token()
    return BlanklineLessNode(nodelist)


@keep_lazy_text
def strip_spaces_html(value):
    return re.sub(r'\s+\\\n\s*', ' ', str(value))


@keep_lazy_text
def strip_spaces_javascript(value):
    value = re.sub(r'\s*\/\/.*\n\s*', '', str(value))
    return re.sub(r'\s*\n\s*', '', value)


class NewlineLessNode(Node):
    def __init__(self, nodelist, mode='html'):
        self.nodelist = nodelist
        self.mode = mode

    def render(self, context):
        if self.mode == 'javascript':
            func = strip_spaces_javascript
        else:
            func = strip_spaces_html
        return func(self.nodelist.render(context).strip())


@register.tag
def newlineless(parser, token):
    """
    Remove escaped newline and whitespace between chars, including tab and
    newline characters.

    Example usage::

        {% newlineless %}
            <a href="foo/" \
               class="bar"> \
                Foo \
            </a>
        {% endnewlineless %}

    This example returns this HTML::

        <a href="foo/" class="bar"> Foo </a>

    """
    nodelist = parser.parse(('endnewlineless',))
    parser.delete_first_token()
    return NewlineLessNode(nodelist)


@register.tag
def newlineless_javascript(parser, token):
    """
    Remove newline and whitespace between chars, including tab and
    newline characters.

    Example usage::

        {% newlineless_javascript %}
            var a = "foo/",
                b = "bar"
            ;
        {% endnewlineless_javascript %}

    This example returns this HTML::

        var a = "foo/",b = "bar";

    """
    nodelist = parser.parse(('endnewlineless_javascript',))
    parser.delete_first_token()
    return NewlineLessNode(nodelist, mode='javascript')


@register.filter
def markdown(text, code_style=None):
    return markdown_to_html(text, code_style)



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#


def get_version(version, repo_dir=None):
    """
    Returns a PEP 386-compliant version number from VERSION.
    """
    version = get_complete_version(version)

    # Now build the two parts of the version number:
    # major = X.Y[.Z]
    # sub = .devN - for pre-alpha releases
    #     | {a|b|c}N - for alpha, beta and rc releases

    major = get_major_version(version)

    sub = ''
    if version[3] == 'alpha' and version[4] == 0:
        changeset = get_vcs_changeset(repo_dir) or '19700101000000'
        sub = '.dev%s' % changeset

    elif version[3] != 'final':
        mapping = {'alpha': 'a', 'beta': 'b', 'rc': 'c'}
        sub = mapping[version[3]] + str(version[4])

    return str(major + sub)


def get_docs_version(version):
    version = get_complete_version(version)
    if version[3] != 'final':
        return 'dev'
    else:
        return '%d.%d' % version[:2]


def get_major_version(version):
    """
    Returns major version from VERSION.
    """
    version = get_complete_version(version)
    parts = 2 if version[2] == 0 else 3
    major = '.'.join(str(x) for x in version[:parts])
    return major


def get_complete_version(version):
    """
    Returns a tuple of the version. If version argument is non-empty,
    then checks for correctness of the tuple provided.
    """
    if not version:
        return (0, 0, 1, 'alpha', 0)
    assert len(version) == 5
    assert version[3] in ('alpha', 'beta', 'rc', 'final')
    return version


def get_vcs_changeset(repo_dir):
    """
    Returns a numeric identifier of the latest VCS changeset.
    """
    from subprocess import Popen, PIPE
    from datetime import datetime

    if not repo_dir:
        from os.path import dirname, abspath
        repo_dir = dirname(dirname(abspath(__file__)))
    commands = (
        'git log --pretty=format:%ct --quiet -1 HEAD',
        'hg tip --template "{date|hgdate}"',
    )
    for command in commands:
        log = Popen(command, stdout=PIPE, stderr=PIPE,
                    shell=True, cwd=repo_dir, universal_newlines=True)
        stamp = log.communicate()[0]
        try:
            stamp = stamp.split(' ')[0]
            stamp = datetime.utcfromtimestamp(int(stamp))
        except ValueError:
            stamp = None
        else:
            break
    return stamp.strftime('%Y%m%d%H%M%S') if stamp else None



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
"""
Functions for access control.
"""
from django.conf import settings

USER_ACCESS_LEVEL = getattr(settings, 'USER_ACCESS_LEVEL', 'user')


def check_user(user):
    """For authenticated users."""
    return user.is_active and user.is_authenticated


def check_employer(user):
    """For employers and superusers."""
    return check_user(user) and (user.is_staff or user.is_superuser)


def check_superuser(user):
    """For superusers."""
    return check_user(user) and user.is_superuser


levels = {
    'anonymous': lambda u: True,
    'user': check_user,
    'employer': check_employer,
    'superuser': check_superuser,
}
# Автоматически выбранная по настройке settings.USER_ACCESS_LEVEL функция.
user_access = levels[USER_ACCESS_LEVEL]



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from os import remove
from os.path import exists

from PIL import Image
from PIL.ExifTags import TAGS


def get_or_create_thumbnail(path, thumb_path, force=False, size=(256, 256),
                            raise_exception=True):
    """
    Создаёт и возвращает объект уменьшенного изображения.
    """
    ex = exists(thumb_path)
    try:
        if force or not ex:
            # (Re)Create thumbnail
            if ex:
                remove(thumb_path)
            img = Image.open(path)
            try:
                exif = img._getexif()
            except Exception:
                exif = None
            if exif is not None:
                for tag, value in exif.items():
                    decoded = TAGS.get(tag, tag)
                    if decoded == 'Orientation':
                        if value == 3:
                            img = img.rotate(180)
                        elif value == 6:
                            img = img.rotate(270)
                        elif value == 8:
                            img = img.rotate(90)
                        break
            rgb = img.convert('RGB')
            rgb.thumbnail(size, Image.ANTIALIAS)
            rgb.save(thumb_path, 'JPEG')
    except IOError as e:
        if raise_exception:
            raise e
        return None
    return thumb_path



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
from django import template
from django.urls import reverse
from django.utils.crypto import get_random_string

register = template.Library()


@register.simple_tag
def active(request, *urls, **kwargs):
    for url in urls:
        # Any a roots does not gets under activity.
        # Namespaces like 'blabla:index' (1 level) does not gets too.
        # But compound, like 'blabla:trololo:index' (2 level and more) gets
        # under activity.
        if (url != '/' and not url.endswith(':index')) or url.count(':') > 1:
            if request.path.startswith(reverse(url, kwargs=kwargs)):
                return "active"
    return ""


@register.simple_tag
def active_equal(request, url, **kwargs):
    if request.path == reverse(url, kwargs=kwargs):
        return "active"
    return ""


@register.simple_tag
def addparams(request, **kwargs):
    q = request.GET.copy()
    for k, v in kwargs.items():
        v = str(v)
        if v:
            q[k] = v
        else:
            q.pop(k, None)
    if q:
        return '?%s' % q.urlencode()
    return ''


@register.simple_tag
def toggleparams(request, **kwargs):
    q = request.GET.copy()
    for k, v in kwargs.items():
        if k in q:
            del q[k]
        else:
            q[k] = str(v)
    if q:
        return '?%s' % q.urlencode()
    return ''


@register.simple_tag
def delparams(request, *params):
    q = request.GET.copy()
    for name in params:
        q.pop(name, None)
    if q:
        return '?%s' % q.urlencode()
    return ''


def page_range_dots(page, on_each_side=3, on_ends=2, dot='.'):
    number = page.number
    paginator = page.paginator
    num_pages = paginator.num_pages
    page_range = paginator.page_range
    if num_pages > 9:
        page_range = []
        if number > (on_each_side + on_ends):
            page_range.extend(range(1, on_each_side))
            page_range.append(dot)
            page_range.extend(range(number + 1 - on_each_side, number + 1))
        else:
            page_range.extend(range(1, number + 1))
        if number < (num_pages - on_each_side - on_ends + 1):
            page_range.extend(range(number + 1, number + on_each_side))
            page_range.append(dot)
            page_range.extend(range(num_pages - on_ends + 1, num_pages + 1))
        else:
            page_range.extend(range(number + 1, num_pages + 1))
    return page_range


@register.filter
def make_page_range(page):
    return page_range_dots(page)


@register.filter
def split(s, sep=','):
    return s.split(sep)


@register.filter
def guid(s, length=12):
    return '%s-%s' % (s, get_random_string(length))



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from django.conf import settings
from django.contrib.auth.decorators import user_passes_test
from django.core.exceptions import PermissionDenied

from djangokit.access.functions import user_access


def user_access_required(function=None, login_url=settings.LOGIN_URL,
                         raise_exception=False, checker=user_access):
    """
    Decorator for views that checks whether a user access.
    """
    def check_perms(user):
        # First, check if the user has permission.
        if checker(user):
            return True
        # If you need a 403 handler, throw an exception.
        if raise_exception:
            raise PermissionDenied
        # Or show the entry form.
        return False
    actual_decorator = user_passes_test(check_perms, login_url=login_url)
    if function:
        return actual_decorator(function)
    return actual_decorator



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from djangokit.utils.deep import to_dict, from_dict


def _save(instance, save):
    if save is not False and hasattr(instance, 'save'):
        fields = set(['details'])
        if save is not True:
            fields = fields.union(save)
        instance.save(update_fields=fields)


class DetailsMixin:
    """Methods for model with JSON field `details`."""
    details = None

    def to_details(self, field, value, append_to_list=False, save=False):
        """
        Advanced setup details parts.

        Default NOT stored in the database. To change this
        behavior, you should pass a parameter `save is not False`.
        """
        self.details = to_dict(
            self.details, field=field, value=value,
            append_to_list=append_to_list,
        )
        _save(self, save)
        return self.details

    def from_details(self, field, default=None, update=False, delete=False,
                     save=False):
        """
        Advanced receiving/removing a portion of the details.

        When receiving can set the specified default value.
        Default NOT stored in the database. To change this
        behavior, you should pass a parameter `save is not False`.
        """
        value = from_dict(
            self.details, field=field, default=default, update=update,
            delete=delete,
        )
        _save(self, save)
        return value

    def set_details(self, details=None, save=False, **kwargs):
        """Simple installation details, or upgrade part details."""
        if details:
            self.details = details
        else:
            d = self.details or dict()
            d.update(kwargs)
            self.details = d
        _save(self, save)
        return self.details


class StateMixin:
    """
    Миксин необходим в первую очередь для проверки нового объекта в шаблонах.
    """

    @property
    def is_new_instance(self):
        return self._state.adding

    @property
    def model_state(self):
        return self._state



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
from django.forms.fields import BooleanField, NullBooleanField
from django.template import Library
from django.utils.encoding import force_text
from django.utils.html import format_html, format_html_join
from django.utils.safestring import mark_safe


register = Library()


@register.filter
def form_group(bound, fgclass=''):
    """
    Formed HTML for visible fields of a forms (boundfield).
    """
    field = bound.field
    if field.__class__.__name__ == 'ReCaptchaField':
        return mark_safe('<div>' + force_text(bound) + '</div>')
    is_nullbool = isinstance(field, NullBooleanField)
    is_checkbox = not is_nullbool and isinstance(field, BooleanField)
    if is_checkbox:
        html = (
            '<div class="form-group %(fgclass)s %(error_class)s">'
            '<div class="checkbox">'
            '<label>%(widget)s %(label)s</label>'
            '%(errors)s'
            '%(help)s'
            '</div>'
            '</div>'
        )
    else:
        html = (
            '<div class="form-group %(fgclass)s %(error_class)s">'
            '%(label)s'
            '%(widget)s'
            '%(errors)s'
            '%(help)s'
            '</div>'
        )
    kw = {'error_class': '', 'errors': '', 'help': '', 'fgclass': fgclass}
    if is_checkbox:
        kw['widget'] = force_text(bound)
        kw['label'] = force_text(bound.label)
    else:
        widget = field.widget
        attrs = widget.attrs
        input_type = getattr(widget, 'input_type', None)
        skip_types = ('hidden', 'file', 'checkbox', 'select', 'radio')
        if 'placeholder' not in attrs and input_type not in skip_types:
            attrs['placeholder'] = force_text(bound.label)
        classes = [c for c in attrs.get('class', '').split(' ') if c]
        if 'form-control' not in classes:
            classes.insert(0, 'form-control')
        attrs['class'] = ' '.join(classes)
        kw['widget'] = bound.as_widget(attrs=attrs)
        label = ''
        if bound.label:
            if field.required:
                label = '<strong>' + bound.label_tag() + '</strong>'
            else:
                label = bound.label_tag()
        kw['label'] = label
    if bound.errors:
        kw['error_class'] = 'text-danger'
        errors = bound.errors.as_data()
        kw['errors'] = format_html(
            '<ul class="text-danger mt-2">{}</ul>',
            format_html_join('', '<li>{}</li>', [e for e in errors])
        )
    if bound.help_text:
        kw['help'] = '<small class="form-text text-muted">%s</small>' % bound.help_text
    return mark_safe(html % kw)



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
"""
Функционал для сериализации данных из моделей.
"""
import logging

from django.db.models import Model, Manager
from django.db.models.fields.files import FieldFile
from django.utils.encoding import force_text

logger = logging.getLogger(__name__)


def _simple_serializer(instance):
    return [
        instance.pk,
        force_text(instance),
    ]


def _complex_serializer(instance):
    return {
        'pk': instance.pk,
        'display_name': force_text(instance),
        # Затем словарь можно дополнить в коде следующим образом:
        # 'fields': {},
        # 'url': reverse(...),
    }


class ObjectSerializer:
    """
    Cериализация полей объекта модели.

    Если передан список полей "only" - то сериализует только их, если его нет,
    то сериализует все поля, за исключением списка "exclude".

    По умолчанию передаётся флаг "simple" и данные формируются в упрощённом
    виде.

    Пример "упрощённой" сериализации:
    ---------------------------------

    post_serializer = ObjectSerializer(exclude=['details'])

    data = post_serializer(post)
    data['url'] = resolve_url(
        'blogs:post', post_furl=post.furl, blog_furl=post.blog.furl,
    )
    # Дополнение поля blog:
    data['blog'][2] = post.blog.furl
    # Дополнение вычисляемых свойств:
    data['count_views'] = post.get_count_views()

    Результатом будет:
    {
        'id': 2,
        'created': '2018-01-01T20:00:00Z',
        'name': 'First post',
        'blog': [1, 'First blog (author: User)', 'first-blog'],
        'count_views': 10,
        'url': '/blogs/first-blog/first-post/',
    }

    Пример "сложной" сериализации:
    ------------------------------

    post_serializer = ObjectSerializer(
        exclude=['blog', 'details'], simple=False
    )
    blog_serializer = ObjectSerializer(only=['name', 'furl'])

    data = post_serializer(post)
    data['url'] = resolve_url(
        'blogs:post', post_furl=post.furl, blog_furl=post.blog.furl,
    )
    # Дополнение поля blog:
    data['fields']['blog'] = blog_serializer(post.blog)
    # Дополнение вычисляемых свойств:
    data['count_views'] = post.get_count_views()

    Результатом будет:
    {
        'display_name': 'First post at 20:00',
        'fields': {
            'id': 2,
            'created': '2018-01-01T20:00:00Z',
            'name': 'First post',
            'blog': {
                'fields': {
                    'furl': 'first-blog',
                    'name': 'First blog',
                }
                'pk': 1,
                'display_name': 'First blog (author: User)',
            }
        },
        'count_views': 10,
        'pk': 2,
        'url': '/blogs/first-blog/first-post/',
    }

    """
    model = None
    only = ()
    exclude = ()
    simple = True

    def __init__(self, model=None, only=None, exclude=None, simple=None):
        if model is not None:
            self.model = model
        if only is not None:
            self.only = only
        elif exclude is not None:
            self.exclude = exclude
        if simple is not None:
            self.simple = simple
        self.is_ready = False
        if self.model:
            self.set_fields(self.model._meta)

    def __call__(self, instance, *args, **kwargs):
        # logger.debug('%s: %s args=%s kwargs=%s', self, instance, args, kwargs)
        return self.serialize(instance, *args, **kwargs)

    def set_fields(self, meta):
        # Собираем все поля в одну кучу.
        all = [f for f in meta.fields]
        if meta.many_to_many:
            all.extend([f for f in meta.many_to_many])
        # Составляем список необходимых полей.
        if self.only:
            list_fields = [f for f in all if f.name in self.only]
        elif self.exclude:
            list_fields = [f for f in all if f.name not in self.exclude]
        else:
            list_fields = all
        self.list_fields = list_fields
        self.is_ready = True
        return list_fields

    def serialize(self, instance, *args, **kwargs):
        """Полная сериализация экземпляра."""
        if not self.is_ready:
            self.set_fields(instance._meta)

        simple = self.simple
        if simple:
            serializer = _simple_serializer
            result = {}
            fields = result
        else:
            serializer = _complex_serializer
            result = serializer(instance)
            fields = result['fields'] = {}

        for f in self.list_fields:
            name = f.name
            value = None
            data = getattr(instance, name, None)
            if isinstance(data, FieldFile):
                if data.name:
                    value = data.url
            elif isinstance(data, Model):
                value = serializer(data)
            elif isinstance(data, Manager):
                value = [serializer(i) for i in data.all()]
            else:
                value = data
            fields[name] = value
        return self.extend(instance=instance, data=result, *args, **kwargs)

    def extend(self, instance, data, *args, **kwargs):
        """Метод для расширения стандартной сериализации."""
        return data



#
# Copyright (c) 2022, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#

from io import StringIO
from pstats import Stats
try:
    from cProfile import Profile
except ImportError:
    from profile import Profile

from django.conf import settings
from django.http import HttpResponse


def profile(get_response):
    """
    В режиме DEBUG возвращает результат работы профилировщика при указании в
    запросе параметра `__profile_view`, либо при наличии в запросе
    заголовка `Profile-View`.
    """

    def check_request(request):
        return '__profile_view' in request.GET or \
            'profile-view' in request.headers

    def middleware(request):
        if settings.DEBUG and check_request(request):
            profiler = Profile()
            profiler.runcall(get_response, request)
            profiler.create_stats()

            io = StringIO()
            stats = Stats(profiler, stream=io)
            stats.print_stats()

            return HttpResponse(io.getvalue())

        return get_response(request)

    return middleware



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
"""A set of extensions for Django used in the construction of complex web
applications.
"""
import os
from djangokit.utils import version

# The version (X, Y, Z, R, N) builds by next rules:
# Variables X, Y, Z & N must be integers. R - can be 'alpha', 'beta' 'rc' or
# 'final' string. R == 'alpha' and N > 0 it is pre-alpha release.
# version = X.Y[.Z]
# subversion = .devN - for pre-alpha releases
#            | {a|b|c}N - for 'alpha', 'beta' and 'rc' releases
# subversion is not exists for 'final' release.
VERSION = (0, 10, 0, 'final', 0)


def get_version():
    path = os.path.dirname(os.path.abspath(__file__))
    return version.get_version(VERSION, path)


__version__ = get_version()



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from datetime import datetime, date, time
from decimal import Decimal
from json import JSONEncoder as OrigJSONEncoder, dumps, loads
from types import GeneratorType
from uuid import UUID

from django.utils.encoding import force_text
from django.utils.functional import Promise
from django.utils.timezone import is_aware


class JSONEncoder(OrigJSONEncoder):
    """
    Подкласс JSONEncoder, который умеет кодировать дату/время, числовой тип,
    генераторы, ленивые объекты перевода и исключения. Почти как в Django, но
    с дополнениями и чуть быстрее.
    """
    use_time_isoformat = True

    def default(self, o):
        # See "Date Time String Format" in the ECMA-262 specification.
        if isinstance(o, datetime):
            r = o.isoformat()
            if not self.use_time_isoformat and o.microsecond:
                r = r[:23] + r[26:]
            if r.endswith('+00:00'):
                r = r[:-6] + 'Z'
            return r
        elif isinstance(o, date):
            return o.isoformat()
        elif isinstance(o, time):
            iso = self.use_time_isoformat
            if not iso and is_aware(o):
                raise ValueError("JSON can't represent timezone-aware times.")
            r = o.isoformat()
            if not iso and o.microsecond:
                r = r[:12]
            if iso and r.endswith('+00:00'):
                r = r[:-6] + 'Z'
            return r
        elif isinstance(o, Decimal):
            return float(o)
        elif isinstance(o, UUID):
            return str(o)
        elif isinstance(o, Exception):
            return force_text(o)
        elif isinstance(o, Promise):
            return force_text(o)
        elif isinstance(o, GeneratorType):
            return list(o)
        else:
            return super(JSONEncoder, self).default(o)


def dump_to_json(data, **kwargs):
    kwargs.setdefault('cls', JSONEncoder)
    return dumps(data, **kwargs)


def dump_from_json(string):
    return loads(string)



#
# Copyright (c) 2017, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from random import choice as random_choice
from uuid import uuid4

from django.utils.crypto import get_random_string
from django.utils.encoding import force_text
from django.utils.functional import Promise
from django.utils.translation import get_language, activate as on_language


def cut_text(obj, cut=255):
    """
    Преобразовывает объект в юникод и обрезает получившуюся строку.
    """
    s = force_text(obj)
    if len(s) > cut:
        return s[:cut]
    return s


def obfuscate_uuid():
    """
    Значение UUID обеспечивает абсолютную 30-летнюю уникальность в 128 бит,
    его обфускация - стойкость в 142 бита (log((6+6+10)**32, 2))
    """
    value = list(uuid4().hex)
    char_indexes = [i for i, char in enumerate(value) if char.isalpha()]
    length = len(char_indexes)
    # Если значение полностью состоит из цифр, то увы, обфускации не выйдет.
    if length > 1:
        random_list = [random_choice([1, 0]) for i in range(length)]
        for i, up in enumerate(random_list):
            if up:
                index = char_indexes[i]
                value[index] = value[index].upper()
    return ''.join(value)


def make_unique_secret(length=40):
    """
    Секретный ключ, с одной стороны, должен быть гарантированно уникальным,
    с другой - обеспечивать достаточную криптостойкость к брутфорсу, но
    при этом не содержать символов пунктуации, так как он может используется
    для авторизации из клиентских приложений.

    Значение UUID обеспечивает абсолютную 30-летнюю уникальность в 128 бит,
    его обфускация - стойкость в 142 бита (log((6+6+10)**32, 2)),
    а рандомный восьмисимвольный суффикс в 47 бит (log((26+26+10)**8, 2))
    - дополнительную гарантию от подбора сектретного ключа.
    """
    assert length >= 40
    return obfuscate_uuid() + get_random_string(length - 32)


def translate_text(lang, text, params=None):
    """
    Принимает текст в виде отложенного перевода и переводит его на язык
    пользователя.

    Все остальные параметры применяются к форматированию текста.

    Примеры использования:
    ----------------------

    >>> text = _('Team "Birds" vs team "Pigs"')
    >>> translate_text('ru', text)
    <<< Команда "Birds" против команды "Pigs"

    либо:

    >>> text = _('Team "%s" vs team "%s"')
    >>> params = [_('Birds'), _('Pigs')]
    >>> translate_text('ru', text, params)
    <<< Команда "Птицы" против команды "Свиньи"

    либо:

    >>> text = _('Team "%(team_1)s" vs team "%(team_2)s"')
    >>> params = {'team_1': _('Birds'), 'team2': _('Pigs')}
    >>> translate_text('ru', text, params)
    <<< Команда "Птицы" против команды "Свиньи"

    """

    def prepare(text, params):
        if isinstance(params, (list, tuple)):
            params = list(params)
            for i, a in enumerate(params):
                if isinstance(a, Promise):
                    params[i] = force_text(a)
            result = force_text(text % tuple(params))
        elif isinstance(params, dict):
            for k, v in params.items():
                if isinstance(v, Promise):
                    params[k] = force_text(v)
            result = force_text(text % params)
        elif params:
            raise ValueError('Params must be list, tuple or dictionary.')
        else:
            result = force_text(text)
        return result

    orig = get_language()
    if orig != lang:
        on_language(lang)
        result = prepare(text, params)
        on_language(orig)
    else:
        result = prepare(text, params)
    return result



#
# Copyright (c) 2017, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from django.http import JsonResponse as BaseJsonResponse
from django.utils.timezone import now

from djangokit.utils.encoders import JSONEncoder


class JsonResponse(BaseJsonResponse):
    def __init__(self, data, **kwargs):
        kwargs.setdefault('encoder', JSONEncoder)
        BaseJsonResponse.__init__(self, data, **kwargs)


class BaseApi(object):
    """Базовый класс для обработки запросов из Node.js."""

    def test(self, request, **kwargs):
        data = {'time': now()}
        data.update(kwargs)
        return data



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
try:
    from django.forms import JSONField as BaseJSONField
except ImportError:
    from django.contrib.postgres.forms.jsonb import JSONField as BaseJSONField
from django.utils.translation import gettext_lazy as _
from djangokit.utils.encoders import dump_to_json


class JSONField(BaseJSONField):
    default_error_messages = {
        'invalid': _('Введите правильный JSON.'),
    }

    def prepare_value(self, value):
        if isinstance(value, str):
            return value
        kwargs = {'indent': 2, 'ensure_ascii': False, 'sort_keys': True}
        return dump_to_json(value, **kwargs)



#
# Copyright (c) 2020, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from os import listdir, stat
from os.path import basename, join, exists
from django.views.generic import View
from djangokit.utils.responses import JsonResponse


class LogReaderView(View):
    """
    AJAX-представление, считывающее построчно log-файлы из определённого
    каталога. Для ограничения доступа
    """
    logdir = None
    maxlines = 1000

    def _tail(self, filename: str, lines: int):
        """Считывает конец файла с некоторым количеством строк перед ним."""
        f = open(filename, 'rb')
        # Чтобы не проходить по очень большим логам, представим, что в среднем
        # строки длиной 1024 байта и перейдём на N килобайт до конца файла.
        avg_size = 1024 * lines
        if stat(filename).st_size > avg_size:
            f.seek(-avg_size, 2)
        buf = []
        i = 0
        for line in f:
            buf.append(line)
            i += 1
            if i // lines == 0:
                buf = buf[-lines:]
        buf = buf[-lines:]
        end = f.tell()
        f.close()
        return {
            'lines': [line.decode('utf-8').rstrip('\n') for line in buf],
            'end': end,
        }

    def _read(self, filename: str, lines: int, start: int):
        """Считывает строки файла начиная с позиции."""
        f = open(filename, 'rb')
        f.seek(start)
        buf = []
        i = 0
        for line in f:
            buf.append(line.decode('utf-8').rstrip('\n'))
            i += 1
            if i >= lines:
                break
        end = f.tell()
        f.close()
        return {
            'lines': buf,
            'end': end,
        }

    def _listdir(self, logdir: str):
        files = [f for f in listdir(logdir) if f.endswith('.log')]
        files.sort()
        return {'files': files}

    def render_to_response(self, context):
        return JsonResponse(context)

    def get(self, request, logfile=None, start=-1, lines=10):
        logdir = self.logdir
        context = {}
        if logfile:
            try:
                start = int(request.GET.get('start', start))
            except ValueError:
                pass
            try:
                lines = int(request.GET.get('lines', lines))
            except ValueError:
                pass
            filename = join(logdir, basename(logfile))
            if exists(filename):
                if lines > self.maxlines:
                    lines = self.maxlines
                if start >= 0:
                    context = self._read(filename, lines, start)
                else:
                    context = self._tail(filename, lines)
        elif logdir and exists(logdir):
            context = self._listdir(logdir)
        return self.render_to_response(context)



from djangokit.access.decorators import user_access_required  # NOQA
from djangokit.access.functions import (
    user_access, check_user, check_employer, check_superuser,
)  # NOQA


__all__ = [
    'user_access', 'check_user', 'check_employer', 'check_superuser',
    'user_access_required',
]



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
"""
Вспомогательный функционал для обработки запросов.
"""
import json
from functools import wraps
from io import BytesIO

from django.http.request import QueryDict, MultiValueDict
from django.http.response import HttpResponseBadRequest
from django.http.multipartparser import MultiPartParserError
from django.utils.encoding import force_text
from django.views.decorators.cache import cache_page


def _user_checksum(user):
    if hasattr(user, 'checksum'):
        return user.checksum
    return str(user.pk)


def cache_user_page(timeout):
    """
    Декоратор кэшированных представлений с разделением на пользователей.
    """

    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):
            user = request.user
            if user.is_authenticated:
                key_prefix = "_auth_%s_" % _user_checksum(user)
            else:
                key_prefix = "_auth_0_"
            view = cache_page(timeout, key_prefix=key_prefix)(view_func)
            return view(request, *args, **kwargs)
        return _wrapped_view
    return decorator


def _file_parser(request):
    if hasattr(request, '_body'):
        # Use already read data
        data = BytesIO(request._body)
    else:
        data = request
    try:
        _data, _files = request.parse_file_upload(request.META, data)
    except MultiPartParserError:
        # An error occurred while parsing POST data. Since when
        # formatting the error the request handler might access
        # request.POST, set request._post and request._file to prevent
        # attempts to parse POST data again.
        # Mark that an error occurred. This allows request.__repr__ to
        # be explicit about it instead of simply representing an
        # empty POST
        request._mark_post_parse_error()
        raise
    return _data, _files


def _form_parser(request):
    data = QueryDict(request.body, encoding=request._encoding)
    return data, MultiValueDict()


def _json_parser(request):
    try:
        data = json.loads(
            force_text(request.body, encoding=request._encoding or 'utf-8') or
            '{}'
        )
        request.json_data_ready = True
    except ValueError:
        request._mark_post_parse_error()
        raise
    return data, MultiValueDict()


def _blank_parser(request):
    data = QueryDict('', encoding=request._encoding)
    return data, MultiValueDict()


def _get_parser(request):
    content_type = request.META.get('CONTENT_TYPE', '')
    if content_type.startswith('multipart/form-data'):
        return _file_parser
    elif content_type.startswith('application/x-www-form-urlencoded'):
        return _form_parser
    elif content_type.startswith('application/json'):
        return _json_parser
    return _blank_parser


def parse_data(function=None, force_to_dict=False):
    """Декоратор для представлений, использующихся в качестве REST API."""

    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):
            request.json_data_ready = False
            if not hasattr(request, 'data'):
                if request.method == 'GET':
                    request.data = request.GET
                else:
                    if request._read_started and not hasattr(request, '_body'):
                        request._mark_post_parse_error()
                        request.data = {}
                    else:
                        parser = _get_parser(request)
                        request.data, request._files = parser(request)
            if force_to_dict:
                data = request.data
                if not isinstance(data, dict):
                    request.data = data.dict()
            return view_func(request, *args, **kwargs)
        return _wrapped_view
    if function:
        return decorator(function)
    return decorator


def json_required(function=None):
    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):
            content_type = request.META.get('CONTENT_TYPE', '')
            if not content_type.startswith('application/json'):
                return HttpResponseBadRequest(
                    'Content type of request must be "application/json".')
            return view_func(request, *args, **kwargs)
        return _wrapped_view
    if function:
        return decorator(function)
    return decorator



#
# Copyright (c) 2016, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from datetime import timedelta
from djangokit.utils.datetimes import datetime_local as now


def gethours(dst=(0, 0), **kwargs):
    """
    Возвращает точки начала и конца часа.
    Параметром dst (кортеж из двух целых чисел) задаётся количество
    часов от текущего для начала и конца периода.
    Наример, если dst == (0, 1) то вернёт:
    (начало-текущего-часа, конец-следующего)
    """
    start = now().replace(minute=0, second=0, microsecond=0)
    end = start.replace(minute=59, second=59, microsecond=999999)
    return (start + timedelta(hours=dst[0]), end + timedelta(hours=dst[1]))


def hour(**kwargs):
    return gethours(dst=(0, 0), **kwargs)


def next_hour(**kwargs):
    return gethours(dst=(1, 1), **kwargs)


def previous_hour(**kwargs):
    return gethours(dst=(-1, -1), **kwargs)


def getdays(dst=(0, 0), withtime=True, **kwargs):
    """
    Возвращает точки начала и конца дня.
    Параметром dst (кортеж из двух целых чисел) задаётся количество
    дней от текущего для начала и конца периода.
    Наример, если dst == (0, 1) то вернёт:
    (начало-текущего-дня, конец-завтрашнего)
    """
    if withtime:
        start = now().replace(hour=0, minute=0, second=0, microsecond=0)
        end = start.replace(hour=23, minute=59, second=59, microsecond=999999)
    else:
        start = end = now().date()
    return (start + timedelta(days=dst[0]), end + timedelta(days=dst[1]))


def today(**kwargs):
    return getdays(dst=(0, 0), **kwargs)


def tomorrow(**kwargs):
    return getdays(dst=(1, 1), **kwargs)


def tomorrow2(**kwargs):
    return getdays(dst=(2, 2), **kwargs)


def yesterday(**kwargs):
    return getdays(dst=(-1, -1), **kwargs)


def yesterday2(**kwargs):
    return getdays(dst=(-2, -2), **kwargs)


def next2days(**kwargs):
    return getdays(dst=(1, 2), **kwargs)


def next3days(**kwargs):
    return getdays(dst=(1, 3), **kwargs)


def last2days(**kwargs):
    return getdays(dst=(-2, -1), **kwargs)


def last3days(**kwargs):
    return getdays(dst=(-3, -1), **kwargs)


def week(dst=0, firstweekday=1, withtime=True, **kwargs):
    """
    Возвращает точки начала и конца недели.
    Параметром dst (целое число) задаётся количество недель от текущей.
    Наример, если dst == 1 то вернёт:
    (начало-следующей-недели, конец-следующей-недели)
    """
    n = now()
    if withtime:
        n = n.replace(hour=0, minute=0, second=0, microsecond=0)
    else:
        n = n.date()

    start = n - timedelta(days=n.weekday() - firstweekday + 1)
    if dst:
        start = start + timedelta(days=(dst * 7))
    if withtime:
        end = start + timedelta(days=7) - timedelta(microseconds=1)
    else:
        end = start + timedelta(days=6)

    return (start, end)


def next_week(**kwargs):
    return week(dst=1, **kwargs)


def previous_week(**kwargs):
    return week(dst=-1, **kwargs)


def get_next_month(dt):
    n_month = ((dt.month + 1) % 12) or 12
    n_year = dt.year + (dt.month // 12)
    return dt.replace(year=n_year, month=n_month, day=1)


def month(dst=0, withtime=True, **kwargs):
    """
    Возвращает точки начала и конца месяца.
    Параметром dst (целое число) задаётся количество месяцев от текущего.
    Наример, если dst == 1 то вернёт:
    (начало-следующго-месяца, конец-следующего-месяца)
    """
    n = now()
    if withtime:
        n = n.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    else:
        n = n.date().replace(day=1)

    month = n.month
    year = n.year

    if dst < 0:
        month = ((n.month - ((abs(dst)) % 12)) % 12) or 12
        year = n.year - ((abs(dst) + 12 - n.month) // 12)
    elif dst > 0:
        month = ((n.month + ((abs(dst)) % 12)) % 12) or 12
        year = n.year + ((abs(dst) + n.month - 1) // 12)

    start = n.replace(year=year, month=month)

    next_month = get_next_month(start)

    if withtime:
        end = next_month - timedelta(microseconds=1)
    else:
        end = next_month - timedelta(days=1)

    return (start, end)


def next_month(**kwargs):
    return month(dst=1, **kwargs)


def previous_month(**kwargs):
    return month(dst=-1, **kwargs)


def quarter1(withtime=True, **kwargs):
    n = now()
    if withtime:
        start = n.replace(
            month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
        end = n.replace(
            month=3, day=31, hour=23, minute=59, second=59, microsecond=999999)
    else:
        n = n.date()
        start = n.replace(month=1, day=1)
        end = n.replace(month=3, day=31)
    return (start, end)


def quarter2(withtime=True, **kwargs):
    n = now()
    if withtime:
        start = n.replace(
            month=4, day=1, hour=0, minute=0, second=0, microsecond=0)
        end = n.replace(
            month=6, day=30, hour=23, minute=59, second=59, microsecond=999999)
    else:
        n = n.date()
        start = n.replace(month=4, day=1)
        end = n.replace(month=6, day=30)
    return (start, end)


def quarter3(withtime=True, **kwargs):
    n = now()
    if withtime:
        start = n.replace(
            month=7, day=1, hour=0, minute=0, second=0, microsecond=0)
        end = n.replace(
            month=9, day=30, hour=23, minute=59, second=59, microsecond=999999)
    else:
        n = n.date()
        start = n.replace(month=7, day=1)
        end = n.replace(month=9, day=30)
    return (start, end)


def quarter4(withtime=True, **kwargs):
    n = now()
    if withtime:
        start = n.replace(
            month=10, day=1, hour=0, minute=0, second=0, microsecond=0)
        end = n.replace(
            month=12, day=31, hour=23, minute=59, second=59, microsecond=999999)
    else:
        n = n.date()
        start = n.replace(month=10, day=1)
        end = n.replace(month=12, day=31)
    return (start, end)


def halfyear1(withtime=True, **kwargs):
    n = now()
    if withtime:
        start = n.replace(
            month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
        end = n.replace(
            month=6, day=30, hour=23, minute=59, second=59, microsecond=999999)
    else:
        n = n.date()
        start = n.replace(month=1, day=1)
        end = n.replace(month=6, day=30)
    return (start, end)


def halfyear2(withtime=True, **kwargs):
    n = now()
    if withtime:
        start = n.replace(
            month=7, day=1, hour=0, minute=0, second=0, microsecond=0)
        end = n.replace(
            month=12, day=31, hour=23, minute=59, second=59, microsecond=999999)
    else:
        n = n.date()
        start = n.replace(month=7, day=1)
        end = n.replace(month=12, day=31)
    return (start, end)


def year(withtime=True, **kwargs):
    n = now()
    if withtime:
        start = n.replace(
            month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
        end = n.replace(
            month=12, day=31, hour=23, minute=59, second=59, microsecond=999999)
    else:
        n = n.date()
        start = n.replace(month=1, day=1)
        end = n.replace(month=12, day=31)
    return (start, end)


def next_year(**kwargs):
    return [x.replace(year=x.year + 1) for x in year(**kwargs)]


def previous_year(**kwargs):
    return [x.replace(year=x.year - 1) for x in year(**kwargs)]



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
"""
Дополнения для django.apps
"""
from django.apps import apps


class DependencesMixin(object):
    """
    Mixin for automatic dependency checking in INSTALLED_APPS.

    Example::

        from django.apps import AppConfig
        from djangokit.utils.apps import DependencesMixin

        class UsersConfig(DependencesMixin, AppConfig):
            name = 'project.users'
            dependences = [
                'django.contrib.sites',
                'django.contrib.staticfiles',
                'django.contrib.sessions',
                'project.core',
            ]

    """
    dependences = None

    def check_dependences(self):
        if self.dependences:
            from django.conf import settings
            from django.core.exceptions import ImproperlyConfigured

            for d in self.dependences:
                if d not in settings.INSTALLED_APPS:
                    raise ImproperlyConfigured(
                        'Application "%s" depends from "%s". But not added '
                        'yet it into INSTALLED_APPS.' % (self.name, d)
                    )

    def ready(self):
        """
        Override this method in subclasses to run code when Django starts.
        """
        self.check_dependences()


def get_dependences_map(package=None):
    configs = apps.get_app_configs()
    L = [(c.name, getattr(c, 'dependences', [])) for c in configs]
    if package:
        packdot = package if package.startswith('.') else package + '.'
        return [x for x in L if x[0] == package or x[0].startswith(packdot)]
    return L



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
"""Crypto functional utilites.
"""
from binascii import b2a_hex, a2b_hex
from Crypto.Random import new as random_new
from Crypto.Cipher.AES import new as cipher_new, block_size, MODE_CBC
from django.utils.encoding import force_bytes, force_text


class AESCipher(object):
    """
    Шифратор по алгоритму AES-128, AES-192 или AES-256 (в зависимости от
    размера "size") с кодированием результата в hex.

    Длина строки данных должна быть кратна 16. Для этого в качестве
    дополнителя используются нечитаемые символы с числом от 1 до 16.
    Например, пустая строка будет заполнена 16-ю '\x10', строка '1234567890'
    превратится в '1234567890\x06\x06\x06\x06\x06\x06'.

    По похожему принципу делаются пароли, но их длина может быть
    16, 24 или 32 символа.

    Заметьте, что для совместимости с Node.js используется MODE_CBC,
    так как другие варианты не обеспечивали расшифровку, хотя может быть
    это была временная ошибка, но так теперь прижилось.

    """

    def __init__(self, password='', size=16):
        assert size in (16, 24, 32), 'Not valid size "%s".' % size
        self.size = size
        self.name = {
            16: 'AES-128-CBC',
            24: 'AES-192-CBC',
            32: 'AES-256-CBC',
        }[size]
        passwd = force_bytes(password)
        length = len(passwd)
        if length > size:
            passwd = passwd[:size]
        elif length < size:
            need = size - (length % size)
            passwd += force_bytes(chr(0) * need)
        self.password = passwd

    def __str__(self):
        return self.name

    def encrypt(self, message):
        """Шифрование данных."""
        IV = random_new().read(block_size)
        aes = cipher_new(self.password, MODE_CBC, IV)
        message = force_bytes(message)
        mod = len(message) % 16
        message += force_bytes((16 - mod) * chr(16 - mod))
        encrypted = b2a_hex(IV) + b2a_hex(aes.encrypt(message))
        return force_text(encrypted)

    def decrypt(self, encrypted):
        """Дешифровка данных."""
        encrypted = a2b_hex(encrypted)
        IV = encrypted[:block_size]
        aes = cipher_new(self.password, MODE_CBC, IV)
        message = aes.decrypt(encrypted[block_size:])
        if message:
            # Python 3 автоматически возвращает число.
            number = message[-1]
            if not isinstance(number, int):
                number = ord(number)
            if number <= 16:
                message = message[:-number]
        else:
            message = ''
        return force_text(message)



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from os.path import dirname as _dirname, basename as _basename
from django import template


register = template.Library()


@register.filter
def dirname(path):
    return _dirname(path)


@register.filter
def basename(path):
    return _basename(path)



#
# Copyright (c) 2016, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
"""Модуль, помогающий обрабатывать глубоко вложенные структуры.
"""
from types import GeneratorType


def split_field(field):
    """Разбивка на поля."""
    if not isinstance(field, (list, tuple)):
        return field.split('.')
    fields = []
    for f in field:
        if not isinstance(f, (list, tuple)):
            f = f.split('.')
        fields.extend(f)
    return fields


def to_dict(dic, field, value, append_to_list=False):
    """
    Рекурсивное обновление поля словаря.

    Ключ значения может выглядеть как:
    'field'
    или
    'field1.field2.field3'
    или
    ['field1', 'field2', 'field3.field4', ...]

    Заметьте, что (по умолчанию) списки поддерживаются только в
    качестве готовых значений! Если нужно добавить в список, то
    передавайте параметр `append_to_list=True`.
    Тогда, если такого списка нет, то он создастся с переданным
    значением внутри, если же есть и это действительно список,
    то добавит в него. Если же назначение не список, то вызовет ошибку.
    """
    D = dic

    if not isinstance(D, dict):
        D = {}

    field = split_field(field)

    d = D
    length = len(field)
    dest = length - 1
    for i in range(0, length):
        key = field[i]
        if key not in d:
            if i == dest:
                if append_to_list:
                    d[key] = [value]
                else:
                    d[key] = value
            else:
                d[key] = {}
                d = d[key]
        elif i == dest:
            if append_to_list:
                d[key].append(value)
            else:
                d[key] = value
        else:
            d = d[key]

    return D


def from_dict(dic, field, default=None, update=False, delete=False):
    """
    Рекурсивное получение поля словаря.

    Ключ значения может выглядеть как:
    'field'
    или
    'field1.field2.field3'
    или
    ['field1', 'field2', 'field3.field4', ...]

    Если задано обновление, то в случае отсутствия значения изменит
    словарь, установив значение по-умолчанию.

    Если задано удаление, то удаляет этот ключ.

    """
    D = dic

    if not isinstance(D, dict):
        D = {}

    field = split_field(field)

    d = D
    value = default
    length = len(field)
    dest = length - 1
    for i in range(0, length):
        key = field[i]
        if key not in d:
            if not update or delete:
                break
            elif i == dest:
                d[key] = value
            else:
                d[key] = {}
                d = d[key]
        elif i == dest:
            if delete:
                value = d.pop(key)
                break
            else:
                value = d[key]
        else:
            d = d[key]

    if callable(value):
        return value()
    else:
        return value


def to_flat_list(data, skip=None, prebool=None, sort=False):
    """
    Преобразовывает многоуровневый объект в [отсортированный] плоский список.
    """
    result = []
    if skip is None:
        skip = ()
    if prebool is None:
        prebool = int

    def collect(item, prefix=''):
        if isinstance(item, dict):
            for key, value in item.items():
                if key in skip:
                    continue
                # Recursive collection for internal values.
                collect(value, prefix='%s%s:' % (prefix, key))
        elif isinstance(item, (list, tuple, range, GeneratorType)):
            for key, value in enumerate(item):
                # Recursive collection for internal values.
                collect(value, prefix='%s%d:' % (prefix, key))
        else:
            if isinstance(item, bool):
                value = prebool(item)
            else:
                value = item
            result.append('%s%s' % (prefix, str(value)))

    collect(data)
    if sort:
        result.sort()
    return result


def to_unique_string(data, skip=None, prebool=None, separator=';'):
    """
    Преобразовывает многоуровневый объект в уникальную строку.
    """
    if data is None:
        return ''
    elif isinstance(data, str):
        return data
    flat_list = to_flat_list(data, skip=skip, prebool=prebool, sort=True)
    return separator.join(flat_list)



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the same license as the current project.
#
from django import template
from djangokit.utils.encoders import dump_to_json

register = template.Library()


@register.filter
def jsondumps(data, indent=None):
    if data is None:
        return ''
    return dump_to_json(data, indent=indent)


@register.filter
def jsondumps_unicode(data, indent=None):
    if data is None:
        return ''
    return dump_to_json(data, indent=indent, ensure_ascii=False)



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
import os
from unidecode import unidecode


def remove_file(filename):
    """Замалчивание ошибки удаления файла."""
    try:
        os.remove(filename)
        return True
    except OSError:
        return False


def remove_dirs(dirname, withfiles=False):
    """Замалчивание ошибки удаления каталога."""
    if withfiles and os.path.exists(dirname):
        for f in os.listdir(dirname):
            remove_file(os.path.join(dirname, f))
    try:
        os.removedirs(dirname)
        return True
    except OSError:
        return False


def secure_filename(s):
    """
    Преобразовывает имена разных алфавитов в латинницу и удаляет/заменяет
    лишние символы.
    """
    return unidecode(s).replace(' ', '_').replace("'", "")



#
# Copyright (c) 2020, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
import logging
from django.core.management.color import color_style


class ColorFormatter(logging.Formatter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.style = color_style(force_color=True)
        self.setup_colorizers()

    def setup_colorizers(self):
        """Установка методов цветового оформления."""
        style = self.style
        self.colorizers = {
            'NOTSET': style.HTTP_SUCCESS,  # без цвета
            'DEBUG': style.HTTP_NOT_MODIFIED,
            'INFO': style.SUCCESS,
            'WARNING': style.WARNING,
            'ERROR': style.ERROR,
            'CRITICAL': style.HTTP_SERVER_ERROR,
        }
        self.default_colorizer = style.HTTP_SUCCESS

    def format(self, record):
        message = super().format(record)
        colorize = self.colorizers.get(record.levelname, self.default_colorizer)
        return colorize(message)



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
"""
Функционал для поиска, сортировки и сериализации данных из моделей и передачи
их представлениям.
"""
from functools import reduce
from logging import getLogger
from operator import or_ as OR

from django.core.exceptions import ValidationError, FieldError
from django.core.paginator import Paginator, EmptyPage
from django.db.models import Q, CharField
from django.forms.forms import BaseForm
from django.shortcuts import redirect, get_object_or_404
from django.shortcuts import _get_queryset
from django.views.generic import TemplateView

from djangokit.serializers import ObjectSerializer


class Controller:
    """
    Filtering, ordering and pagination controller for QuerySet.
    """
    model = None
    form = None
    search_fields = None
    ordering_fields = None
    filtering_fields = None
    exclude_fields = ('password',)
    select_related = None
    prefetch_related = None
    serializer = None
    use_distinct = False
    default_ordering = None

    def __init__(self, model=None, search_fields=None, ordering_fields=None,
                 filtering_fields=None, exclude_fields=None, serializer=None,
                 select_related=None, prefetch_related=None,
                 default_ordering=None):
        if model:
            self.model = model
        else:
            model = self.model
        meta = model._meta
        if search_fields is not None:
            self.search_fields = search_fields
        if ordering_fields is not None:
            self.ordering_fields = ordering_fields
        if filtering_fields is not None:
            self.filtering_fields = filtering_fields
        if exclude_fields is not None:
            self.exclude_fields = exclude_fields
        if select_related is not None:
            self.select_related = select_related
        if prefetch_related is not None:
            self.prefetch_related = prefetch_related
        exclude = self.exclude_fields
        fields = [f for f in meta.fields if f.name not in exclude]
        if meta.many_to_many:
            fields.extend([
                f for f in meta.many_to_many if f.name not in exclude
            ])
        self.fields = fields
        if self.serializer is None:
            self.serializer = (
                serializer or ObjectSerializer(model=model, exclude=exclude)
            )
        if isinstance(self.serializer, type):
            self.serializer = self.serializer(model=model, exclude=exclude)
        # Устанавливаем логгер.
        if self.__class__.__module__ != __name__:
            orig_name = 'djangokit.Controller.' + str(self)
        else:
            orig_name = 'djangokit.Controller'
        self.logger = getLogger(orig_name + '(' + str(meta) + ')')
        # Устанавливаем дефолтную сортировку.
        if default_ordering is not None:
            self.default_ordering = default_ordering
        elif not self.default_ordering:
            self.default_ordering = meta.ordering or ('pk',)

    def __str__(self):
        cls = self.__class__
        return cls.__module__ + '.' + cls.__name__

    def get_queryset(self):
        qs = _get_queryset(self.model)
        if self.select_related:
            qs = qs.select_related(*self.select_related)
        if self.prefetch_related:
            qs = qs.prefetch_related(*self.prefetch_related)
        return qs

    def get_search_fields(self):
        if self.search_fields is None:
            fields = [f.name for f in self.fields if isinstance(f, CharField)]
            # Search by models without text fields.
            if not fields:
                exclude = self.exclude_fields
                for field in self.fields:
                    rel = field.related_model
                    if rel:
                        prefix = field.name + '__%s'
                        for f in rel._meta.fields:
                            if isinstance(f, CharField):
                                fname = prefix % f.name
                                if fname not in exclude:
                                    fields.append(fname)
            self.search_fields = fields
        return self.search_fields

    def get_ordering_fields(self):
        if self.ordering_fields is None:
            self.ordering_fields = [f.name for f in self.fields if
                                    not f.related_model]
        return self.ordering_fields

    def get_filtering_fields(self):
        if self.filtering_fields is None:
            self.filtering_fields = [f.name for f in self.fields]
            self.filtering_fields += [
                rel.name for rel in self.model._meta.related_objects
            ]
        return self.filtering_fields

    def ordering(self, queryset, ordering):
        """
        Функция проверяет параметры сортировки и применяет только допустимые.
        """
        logger = self.logger
        fields = self.get_ordering_fields()
        if not fields:
            logger.debug('%s has no fields for ordering', self)

        if not ordering or not fields:
            # Fix UnorderedObjectListWarning:
            if not getattr(queryset, 'ordered', False):
                queryset = queryset.order_by(*self.default_ordering)
                logger.debug('default ordering by %s', self.default_ordering)
            return queryset, []

        def valid(x):
            v = bool(
                x and not x.startswith('--') and
                x.lstrip('-') in fields
            )
            if not v:
                logger.debug('invalid ordering field: %s', x)
            return v

        if isinstance(ordering, str):
            if ordering.startswith('[') and ordering.endswith(']'):
                ordering = ordering[1:-1]
            ordering = ordering.split(',')
        ordering = [x for x in ordering if valid(x)]
        if ordering:
            queryset = queryset.order_by(*ordering)
            logger.debug('ordering by %s', ordering)
        # Fix UnorderedObjectListWarning:
        elif not getattr(queryset, 'ordered', False):
            queryset = queryset.order_by(*self.default_ordering)
            logger.debug('default ordering by %s', self.default_ordering)
        return queryset, ordering

    def search(self, queryset, query):
        """
        Фильтрует набор данных поиском по запросу.
        """
        logger = self.logger

        fields = self.get_search_fields()

        def construct_search(field_name):
            if field_name.startswith('^'):
                return "%s__istartswith" % field_name[1:]
            elif field_name.startswith('='):
                return "%s__iexact" % field_name[1:]
            elif field_name.startswith('@'):
                return "%s__search" % field_name[1:]
            else:
                return "%s__icontains" % field_name

        if fields and query not in ('', None, False, True):
            lookups = [construct_search(str(f)) for f in fields]
            for bit in query.split():
                queries = [Q(**{lookup: bit}) for lookup in lookups]
                queryset = queryset.filter(reduce(OR, queries))
            logger.debug('search "%s" by %s', query, lookups)

        return queryset

    def filtering(self, queryset, filters):
        """Фильтрует набор данных."""
        logger = self.logger

        if not filters:
            return queryset, {}

        fields = self.get_filtering_fields()

        def test_filtered(field):
            return field.split('__')[0] in fields

        def test_inverse(s):
            return s.startswith('-')

        def test_bool(s, v):
            return s.endswith('__isnull') or v in ('true', 'false')

        def test_list(s):
            return s.endswith('__in') or s.endswith('__range')

        applied = {}

        for field, query in filters.items():
            if field == 'q':
                queryset = self.search(queryset, query)
                applied[field] = query
                continue

            if test_inverse(field):
                field = field[1:]
                func = queryset.exclude
            else:
                func = queryset.filter

            if not test_filtered(field):
                logger.debug('field %s not filtered', field)
                continue

            if isinstance(query, str):
                if query.startswith('[') and query.endswith(']'):
                    query = [x for x in query[1:-1].split(',') if x]
                    if not query:
                        continue
                elif test_list(field):
                    query = [x for x in query.split(',') if x]
                    if not query:
                        continue
                elif query == 'null':
                    query = None
                elif test_bool(field, query):
                    query = bool(query == 'true')
            try:
                queryset = func(Q(**{field: query}))
            except (ValueError, ValidationError, FieldError):
                pass
            else:
                applied[field] = query

        return queryset, applied

    def pagination(self, queryset, limit, page):
        """Возвращает пагинатор и страницу для набора данных."""
        paginator = Paginator(queryset, limit)
        try:
            page = paginator.page(page)
        except EmptyPage:
            self.logger.debug('page %s is empty', page)
            page = paginator.page(1)
        return paginator, page

    def distinct(self, queryset):
        query = queryset.query
        fields = query.extra_order_by or query.order_by
        meta = query.get_meta()
        if not fields and query.default_ordering:
            fields = meta.ordering
        fields = [f.lstrip('-') for f in fields]
        pk = meta.pk.name
        fields = [f for f in fields if f != pk]
        queryset = queryset.distinct(pk, *fields)
        self.logger.debug('distinct by %s', fields)
        return queryset

    def make_form(self, request):
        return self.form(request.GET)

    def get(self, request, queryset=None, page=1, limit=100, max_limit=1000):
        """
        Подготавливает запрос и возвращает объект страницы пагинатора,
        актуальную сортировку и применённые фильтры.
        """
        logger = self.logger
        if self.form:
            form = self.make_form(request)
            form.is_valid()
            data = form.cleaned_data
        else:
            form = None
            data = request.GET.dict()
        if queryset is None:
            queryset = self.get_queryset()
        else:
            assert self.model == queryset.model
        if 'p' in data:
            _p = data.pop('p')
            try:
                page = int(_p)
            except ValueError:
                logger.debug('invalid page "%s" in request', _p)
                pass
        if 'l' in data:
            _l = data.pop('l')
            try:
                limit = int(_l)
            except ValueError:
                logger.debug('invalid limit "%s" in request', _l)
                pass
        if limit > max_limit:
            logger.debug('limit "%s" changed to max "%s"', limit, max_limit)
            limit = max_limit
        queryset, orders = self.ordering(queryset, data.pop('o', None))
        if self.use_distinct:
            queryset = self.distinct(queryset)
        queryset, filters = self.filtering(queryset, data)
        logger.debug('%s', queryset.query)
        paginator, page = self.pagination(queryset, limit, page)
        if form:
            page.form = form
        return page, orders, filters

    def get_serialized(self, *args, **kwargs):
        serializer = kwargs.pop('serializer', None)
        page, orders, filters = self.get(*args, **kwargs)
        page = self.serialize_page(page, serializer)
        return page, orders, filters

    def serialize_page(self, page, serializer):
        if serializer is None:
            serializer = self.serializer
        paginator = page.paginator
        return {
            'objects': [serializer(obj) for obj in page.object_list],
            'number': page.number,
            'limit': paginator.per_page,
            'count': paginator.count,
            'pages': paginator.num_pages,
        }


class ControlModelView(TemplateView):
    template_name = None
    ctrl = None
    form = None
    redirect_after_add = None
    limit = 100
    max_limit = 1000

    def __init__(self, *args, **kwargs):
        ctrl = self.ctrl
        assert isinstance(ctrl, Controller)
        self.model = ctrl.model
        assert issubclass(self.form, BaseForm)
        return super().__init__(*args, **kwargs)

    def get_default_context(self):
        return {}

    def add_form_to_context(self, request, instance, context):
        Form = self.form
        kw = {'instance': instance}
        if hasattr(Form, 'request'):
            kw['request'] = request
        if request.method == 'POST':
            kw['files'] = request.FILES
            context['form'] = form = Form(request.POST, **kw)
            if form.is_valid():
                context['instance'] = form.save()
        else:
            context['form'] = Form(**kw)
        return context

    def get_allowed_queryset(self, request, for_instance):
        return self.ctrl.get_queryset()

    def make_context(self, request, id):
        ctx = self.get_default_context()
        extra = self.extra_context
        if extra:
            ctx.update(extra)

        queryset = self.get_allowed_queryset(request, for_instance=id)
        if id:
            instance = get_object_or_404(queryset, id=id)
        else:
            instance = None
            page, orders, filters = self.ctrl.get(
                request, limit=self.limit, max_limit=self.max_limit,
                queryset=queryset,
            )
            ctx.update({
                'page': page,
                'orders': orders,
                'orders_string': ','.join(orders) if orders else '',
                'filters': filters,
            })
            ctx['next'] = request.path
        ctx['model'] = self.model
        ctx['ctrl'] = self.ctrl
        ctx['instance'] = instance
        self.add_form_to_context(request, instance, ctx)
        return ctx

    def get(self, request, id=None):
        ctx = self.make_context(request, id)
        if 'next' in request.GET:
            ctx['next'] = request.GET['next']
        return self.render_to_response(ctx)

    def post(self, request, id=None):
        ctx = self.make_context(request, id)
        form = ctx['form']
        is_valid = form.is_bound and not form._errors
        if not is_valid:
            return self.render_to_response(ctx)
        to = request.POST.get('next')
        if to:
            return redirect(to)
        instance = ctx.get('instance')
        if instance and id is None and self.redirect_after_add:
            return redirect(self.redirect_after_add, id=instance.id)
        if instance and hasattr(instance, 'get_absolute_url'):
            return redirect(instance.get_absolute_url())
        return self.render_to_response(ctx)



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from setuptools import setup

# Dynamically calculate the version based on djangokit.VERSION.
version = __import__('djangokit').get_version()

with open("README.rst", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    requirements = [
        line.split('#', 1)[0].strip() for line in f.read().splitlines()
        if not line.strip().startswith('#')
    ]

with open('requirements-optional.txt') as f:
    optional_requirements = [
        line.split('#', 1)[0].strip() for line in f.read().splitlines()
        if not line.strip().startswith('#')
    ]


setup(
    name="DjangoKit",
    version=version,
    author="Grigoriy Kramarenko",
    author_email="root@rosix.ru",
    description=("DjangoKit is a set of extensions for Django."),
    long_description=long_description,
    url="https://gitlab.com/djbaldey/djangokit/",
    license="BSD License",
    zip_safe=False,
    packages=['djangokit'],
    include_package_data=True,
    install_requires=requirements,
    extras_require={
        'full': optional_requirements,
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Natural Language :: Russian",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.5",
)



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
try:
    from markdown import markdown
    from pygments.styles import STYLE_MAP
except ImportError:
    use_markdown = False
    STYLE_MAP = {}
else:
    use_markdown = True

from django.utils.safestring import mark_safe

_html_escapes = {
    ord('&'): '&amp;',
    ord('<'): '&lt;',
    ord('>'): '&gt;',
    # ord('"'): '&quot;',
    # ord("'"): '&#39;',
}


def _markdown(source, code_style):
    if not use_markdown:
        return source
    if not code_style or code_style not in STYLE_MAP:
        code_style = 'default'
    conf = {
        'codehilite': {
            'linenums': False,
            'css_class': 'text-monospace codehilite',
            'noclasses': True,
            'pygments_style': code_style,
        }
    }
    extensions = [
        'attr_list',
        'fenced_code',
        'codehilite',
        'tables',
    ]
    return markdown(source, extensions=extensions, extension_configs=conf)


def markdown_to_html(source, code_style=None):
    """
    Обработка Markdown с подстветкой синтаксиса через Pygments для блоков кода
    в стиле GitHub-Flavored Markdown.
    """
    text = []
    code_started = False

    for s in source.split('\n'):
        start_gfm_code = s.strip().startswith('```')
        if start_gfm_code and not code_started:
            text.append(s)
            code_started = True
        elif start_gfm_code and code_started:
            text.append(s)
            code_started = False
        elif code_started:
            text.append(s)
        else:
            # Всё остальное нужно обработать от внедрения кода.
            text.append(mark_safe(str(s).translate(_html_escapes)))
    prepared = '\n'.join(text)
    return _markdown(prepared, code_style)



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from django.conf import settings
from django.shortcuts import render, redirect


class MaintenanceMiddleware:
    """
    Middleware постановки сайта на обслуживание.

    Если в настройках включена настройка MAINTENANCE,
    на все запросы отдает страницу maintenance.html c кодом 404, чтобы
    сообщить, что этот документ в ближайшее время будет восстановлен.

    Шаблон и код ответа модно переопределить через начледование класса.
    """

    template_name = 'maintenance.html'
    status_code = 404
    on = getattr(settings, 'MAINTENANCE', False)

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if not self.on:
            return self.get_response(request)
        return render(request, self.template_name, status=self.status_code)


class SeparateAccessMiddleware:
    """
    Middleware для выборочного доступа ко всем ресурсам сайтов.

    Если в настройках включена настройка USE_SEPARATE_ACCESS,
    то каждый запрос проверяется на выборочный доступ, и если он не
    проходит, то на все запросы отдает страницу separate_access.html и код 403.
    На данный странице предлагается ввести ключ, который отредиректит
    строго на главную страницу.

    Внимание! Не добавляйте эту прослойку в административный сайт, если
    используете параметр, получаемый из своей модели параметров в базе данных
    через кастомизированный метод `get_key_from_db`. Да, этот метод нужен
    именно для переопределения в своих проектах.
    """
    template_name = 'separate_access.html'
    status_code = 403
    separate_access_key = getattr(settings, 'SEPARATE_ACCESS_KEY', '')
    session_key_name = 'access_key'
    post_key_name = 'access_key'

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Без ключей это должно работать быстро. Поэтому сначала
        # просто проверяем наличие ограничения.
        separate_access_key = self.separate_access_key
        if not separate_access_key:
            return self.get_response(request)
        _key = self.get_key_from_db(request, separate_access_key)
        # Затем проверяем ключ из сессии на актуальность.
        session_key_name = self.session_key_name
        session = request.session
        access_key = session.get(session_key_name)
        if access_key == _key:
            return self.get_response(request)
        elif access_key:
            del session[session_key_name]
        # Проверка ключа доступа в параметрах запроса.
        if request.method == 'POST':
            access_key = request.POST.get(self.post_key_name)
            if access_key == _key:
                session[session_key_name] = access_key
                return redirect(request.path)
        return render(request, self.template_name, status=self.status_code)

    def get_key_from_db(self, request, default):
        return default



#
# Copyright (c) 2019, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from django import forms
from django.http import QueryDict
from django.shortcuts import resolve_url


class SelectizeMixin(object):
    # refurl_template_name = ''

    def __init__(self, refurl=None, filters=None, create=False, *args, **kwargs):
        self.refurl = refurl
        self.filters = filters
        self.has_create_object = create
        if refurl:
            self.template_name = self.refurl_template_name
        super().__init__(*args, **kwargs)

    def get_context(self, name, value, attrs):
        classes = [c for c in attrs.get('class', '').split(' ') if c]
        if 'selectize-is-not-set' not in classes:
            classes.append('selectize-is-not-set')
        attrs['class'] = ' '.join(classes)
        if self.has_create_object:
            attrs['data-create'] = 'true'
        # Собираем контент для нашего шаблона.
        if self.refurl:
            return self.get_refurl_context(name, value, attrs)
        # Отдаём контент для встроенного шаблона.
        return super().get_context(name, value, attrs)

    def get_refurl_context(self, name, value, attrs):
        attrs['data-url'] = resolve_url(self.refurl)
        if self.filters:
            query = QueryDict('', mutable=True)
            for k, v in self.filters.items():
                if isinstance(v, (tuple, list)):
                    v = ','.join(v)
                query.appendlist(k, v)
            query._mutable = False
            attrs['data-filters'] = query.urlencode()
        if isinstance(value, (tuple, list)):
            value = [x for x in value if x]
        context = {}
        context['widget'] = {
            'name': name,
            'is_hidden': self.is_hidden,
            'required': self.is_required,
            'value': self.format_value(value),
            'attrs': self.build_attrs(self.attrs, attrs),
            'template_name': self.template_name,
        }
        if self.allow_multiple_selected:
            context['widget']['attrs']['multiple'] = True
        if value:
            context['value'] = self.get_instance(value)
        return context

    def get_instance(self, value):
        qs = self.choices.queryset
        if isinstance(value, (tuple, list)):
            return qs.filter(pk__in=value)
        return qs.get(pk=value)


class Selectize(SelectizeMixin, forms.Select):
    refurl_template_name = 'djangokit/widgets/selectize.html'


class SelectizeMultiple(SelectizeMixin, forms.SelectMultiple):
    refurl_template_name = 'djangokit/widgets/selectize_multiple.html'



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
from functools import wraps
from os import remove
from os.path import exists

from django.utils.dateparse import parse_datetime
from django.utils.timezone import now, timedelta


class LockError(Exception):
    pass


def lockfile(method=None, filename=None, timeout=60 * 60):
    """
    Декоратор для метода handler команды вызов которой блокируется файлом.
    По умолчанию: название файла берётся по имени команды, блокировка на 1 час.
    """

    def get_filename(command):
        if not filename:
            return '%s.lock' % command.__class__.__module__.split('.')[-1]
        return filename

    def get_timeout():
        return timeout

    def decorator(method_func):
        @wraps(method_func)
        def _wrapped_method(self, *args, **kwargs):
            filename = get_filename(self)
            timeout = get_timeout()
            # TODO: UNIX only, please fix it.
            path = '/tmp/%s' % filename
            newtime = now()
            if exists(path):
                with open(path, 'r') as f:
                    line = f.readline()
                    time = parse_datetime(line)
                    if time and time + timedelta(seconds=timeout) > newtime:
                        raise LockError(
                            'Blocked by "%s" file at "%s"' % (path, time)
                        )
            f = open(path, 'w')
            f.write(newtime.isoformat())
            f.close()
            try:
                return method_func(self, *args, **kwargs)
            except Exception as e:
                raise e
            finally:
                remove(path)
        return _wrapped_method

    if method:
        return decorator(method)
    return decorator



#
# Copyright (c) 2018, Grigoriy Kramarenko
# All rights reserved.
# This file is distributed under the BSD 3-Clause License.
#
import traceback
from functools import wraps
from logging import getLogger

from django.utils.encoding import force_text
from django.utils.log import AdminEmailHandler as BaseAdminEmailHandler


class AdminErrorHandler(BaseAdminEmailHandler):
    """Обработчик логгера запускаемых команд, отправляющий почту админам."""

    def emit(self, record):
        subject = message = None

        if hasattr(record, 'command') and hasattr(record, 'exception'):
            command = record.command
            name = command.__class__.__module__.split('.')[-1]
            subject = 'Something went wrong in %s' % name

            exception = record.exception
            message = ''.join(traceback.format_exception(
                etype=type(exception),
                value=exception,
                tb=exception.__traceback__,
            ))
            if hasattr(record, 'command_args'):
                message += '\n\nARGS:\n    %s' % '\n    '.join(
                    [force_text(i) for i in record.command_args]
                )
            if hasattr(record, 'command_kwargs'):
                message += '\n\nKWARGS:\n    %s' % '\n    '.join(
                    ['%s=%s' % (force_text(k), force_text(v)) for k, v in
                     record.command_kwargs.items()]
                )
        else:
            subject = '%s: %s' % (
                record.levelname,
                record.getMessage()
            )
            subject = self.format_subject(subject)
            message = 'Undefined message, please check server logs.'

        self.send_mail(subject, message, fail_silently=True)


def handler_logging(method=None, level='error', logger_or_name='management'):
    """
    Декоратор для метода handler команды, который отправляет логгеру
    ошибки исполнения.
    """

    # Проверяем наличие метода в момент определения кода.
    assert level in ('debug', 'info', 'warning', 'error', 'critical')
    if isinstance(logger_or_name, str):
        logger = getLogger(logger_or_name)
    else:
        logger = logger_or_name
    log_func = getattr(logger, level)

    def decorator(method_func):
        @wraps(method_func)
        def _wrapped_method(self, *args, **kwargs):
            try:
                return method_func(self, *args, **kwargs)
            except Exception as e:
                log_func(
                    force_text(e),
                    extra={
                        'command': self,
                        'command_args': args,
                        'command_kwargs': kwargs,
                        'exception': e,
                    }
                )
                raise e
        return _wrapped_method

    if method:
        return decorator(method)
    return decorator
